from typing import Any, Generator, List, Sequence
from Bio.Align import PairwiseAligner
from Bio import Entrez
from Bio import SeqIO

from mlstmyfasta.engine.data.genomics import Strand, StrandFeature, get_feature_coding


async def fetch_ncbi_genbank(genbank_id: str) -> Strand:
    with Entrez.efetch(db="nucleotide", id=genbank_id, rettype="gb", retmode="text") as fetch_stream:
        record = SeqIO.read(fetch_stream, "genbank")
        sequence_features = list()
        for feature in record.features:
            start = int(feature.location.start)
            end = int(feature.location.end)
            sequence_features.append(StrandFeature(
                type=feature.type,
                start=start,
                end=end+1,  # Position is exclusive
                feature_properties=feature.qualifiers
            ))
        return Strand(name=genbank_id, coding=str(record.seq), features=sequence_features)


async def annotate_from_genbank(genbank_id: str, query_coding: str):
    strand = await fetch_ncbi_genbank(genbank_id=genbank_id)
    aligner = PairwiseAligner("blastn")
    aligner.mode = "local"
    for feature in strand.features:
        feature_coding = get_feature_coding(strand=strand, feature=feature)
        if len(aligner.align(query_coding, feature)) < 1:
            # TODO implement a failsafe
            continue
        sequence = sorted(aligner.align(query_coding, feature))[0]
